/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *          
 */

#include <config.h>

#include <string.h>
#include <math.h>

#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>
#include <gdk/gdkkeysyms.h>

#include "im-ja.h"
#include "error.h"
#include "conf.h"
#include "common.h"
#include "nls.h"
#include "candwin.h"
#include "statuswin.h"
#include "kanjipad.h"

#ifndef DISABLE_CANNA
#include "canna_rk.h"
#endif
#ifndef DISABLE_WNN
#include "wnn.h"
#endif

extern IMJAConfig cfg; 

static void candidate_window_set_selection(GtkWidget *menu_item, IMContextJA *cn) {
	gchar *string;
	GtkWidget *label;

	IM_JA_DEBUG("candidate_window_set_selection()\n");
	
	//get the label of the menuitem
	label = gtk_container_get_children(GTK_CONTAINER(menu_item))->data;
	string = (gchar *)gtk_label_get_text(GTK_LABEL(label));

	gint selected = (gint) g_object_get_data(G_OBJECT(menu_item), "candidate-number");
	if (cn->conv_engine == CONV_ENGINE_CANNA) {
#ifndef DISABLE_CANNA
		IM_JA_DEBUG("Candidate window calling canna_rk_choose_candidate(%d)\n",selected);
		canna_rk_choose_candidate(cn, selected);
#endif
	}
	else if (cn->conv_engine == CONV_ENGINE_WNN) {
#ifndef DISABLE_WNN
		im_ja_wnn_select_candidate(cn, selected);
		im_ja_wnn_update_preedit(cn);
#endif
	}
	else {
		g_signal_emit_by_name(cn, "commit", string);
	}
	candidate_window_hide(cn);
	status_window_show(cn);
}

static void candidate_window_change_selection(GtkWidget *menu_item, IMContextJA *cn) {
	gint selected = 0;
	//GList *candidate = cn->candidate_list;
	gchar *new_preedit;
	GtkWidget *label;
	gchar *string;
	
	//get the label of the menuitem
	label = gtk_container_get_children(GTK_CONTAINER(menu_item))->data;
	string = (gchar *)gtk_label_get_text(GTK_LABEL(label));

	selected = (gint) g_object_get_data(G_OBJECT(menu_item), "candidate-number");
	if (cn->conv_engine == CONV_ENGINE_WNN) {
#ifndef DISABLE_WNN
		im_ja_wnn_select_candidate(cn, selected);
		im_ja_wnn_update_preedit(cn);
#endif
	}
	else {
		new_preedit = g_new0(guchar, BUFFERSIZE);
		strncpy(new_preedit, cn->preedit_buf, cn->preedit_reverse_start);
		g_strlcat(new_preedit, string, BUFFERSIZE);
		g_strlcat(new_preedit, cn->preedit_buf + cn->preedit_reverse_end, BUFFERSIZE);
		cn->preedit_reverse_end = strlen(string);
		g_strlcpy(cn->preedit_buf, new_preedit, BUFFERSIZE);
		cn->preedit_reverse_end = cn->preedit_reverse_start + strlen(string);
		g_free(new_preedit);
		g_signal_emit_by_name(cn, "preedit_changed");
	}
	
}

static void candidate_window_set_position(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, IMContextJA *cn) {
	gint cw_x = 0, cw_y = 0; //client pos
	gint target_x = 0, target_y = 0;
	GdkRectangle toplevel_rect;
		
	if (cn->toplevel_gtk != NULL) {
		gdk_window_get_frame_extents(cn->toplevel_gtk->window, &toplevel_rect);
		if (cn->client_gdk != NULL) gdk_window_get_origin(cn->client_gdk, &cw_x, &cw_y);
	}
	else {
		*x = 0;
		*y = 0;
		return;
	}

	if ((cn->candwin_pos_offset_x == 0) && (cn->candwin_pos_offset_y == 0)) {
		//If we don't receive the position properly, center on client window
		GtkRequisition requisition;
		gtk_widget_size_request(GTK_WIDGET(menu), &requisition);
		gdk_window_get_frame_extents(cn->toplevel_gtk->window, &toplevel_rect);
		*x = toplevel_rect.x + toplevel_rect.width / 2 - requisition.width / 2;
		*y = toplevel_rect.y + toplevel_rect.height / 2 - requisition.height / 2;
	}
	else {
		target_x = cw_x + cn->candwin_pos_offset_x;
		target_y = cw_y + cn->candwin_pos_offset_y;
		/* check if target_ is within the toplevel window */
		if ((toplevel_rect.width != 0) || (toplevel_rect.height != 0)) {
			if (target_y > toplevel_rect.y + toplevel_rect.height) target_y = toplevel_rect.y + toplevel_rect.height;
			if (target_x > toplevel_rect.x + toplevel_rect.width) target_x = toplevel_rect.x + toplevel_rect.width;
		}
		*x = target_x;
		*y = target_y + 12; //FIXME: this constant (12) should be fixed. Use font height or something.
	}

	IM_JA_DEBUG("candwin pos x: %d, y: %d\n", *x, *y);
}

static gboolean key_press_cb(GtkWidget *widget, GdkEventKey *event, IMContextJA* cn) {
	gint ignorekeys[] = { GDK_Up, GDK_KP_Up, GDK_Down, GDK_KP_Down, GDK_KP_Left, 
												GDK_KP_Right, GDK_Home, GDK_KP_Home, GDK_End, GDK_KP_End, GDK_Page_Up,
												GDK_KP_Page_Up, GDK_Page_Down, GDK_KP_Page_Down, GDK_Return };
	gint i;
	gboolean found = FALSE;
	gint size = sizeof(ignorekeys)/sizeof(gint);
	IM_JA_DEBUG("candidate_win key_press_cb()\n");

	if (cfg.candwin_style == CANDWIN_STYLE_TABLE) {
		if (event->keyval == GDK_Escape) {
			candidate_window_hide(cn);
			return TRUE;
		}
		return FALSE;
	}
	else { //CANDWIN_STYLE_MENU
		// Select the next candidate if space was pressed
		if (event->keyval == GDK_space) {
			GList *node = g_list_find(GTK_MENU_SHELL(cn->candidate_win->window)->children,
																GTK_MENU_SHELL(cn->candidate_win->window)->active_menu_item);
			node = node->next;
			if (node == NULL) node = GTK_MENU_SHELL(cn->candidate_win->window)->children;
			gtk_menu_shell_select_item(GTK_MENU_SHELL(cn->candidate_win->window), node->data);
			return TRUE;
		}
	
		/* Handle next/prev bunsetu selection */
		if ((ishotkey(event, NEXT_BUNSETSU, &cfg) == TRUE) ||
				(ishotkey(event, PREV_BUNSETSU, &cfg) == TRUE) ||
				(event->keyval == GDK_Left) || (event->keyval == GDK_Right)) {
			GList *node = g_list_find(GTK_MENU_SHELL(cn->candidate_win->window)->children,
																GTK_MENU_SHELL(cn->candidate_win->window)->active_menu_item);
			candidate_window_set_selection(node->data, cn);
			candidate_window_hide(cn);
			return gtk_im_context_filter_keypress(GTK_IM_CONTEXT(cn), event);
		}

		/* Check if this key is handled by the menu or not */
		for (i = 0; i < size; i++) {
			if (ignorekeys[i] == event->keyval) {
				found = TRUE;
				break;
			}
		}

		/* Only process keys that the menu doesn't use for setting it's selections */
		if (found == FALSE) {
			IM_JA_DEBUG("key press in candidate menu.\n");
			candidate_window_hide(cn);
			return gtk_im_context_filter_keypress(GTK_IM_CONTEXT(cn), event); //Maybe we should commit ?
		}
		else return FALSE;
	}
}

/*
static void candidate_window_cancel_cb(IMContextJA* cn) {
	IM_JA_DEBUG("candidate_window_cancel_cb()\n");

	if (selection_ok == TRUE) return;
	if (cn->candidate_win == NULL) return;
	if (GTK_IS_WIDGET(cn->candidate_win->window) == FALSE) return;

	if (cn->conv_engine == CONV_ENGINE_CANNA) {
#ifndef DISABLE_CANNA
		IM_JA_DEBUG("FIXME: candidate_window_cancel_cb for canna\n");
		//canna_rk_choose_candidate(cn, selected);
#endif
	}
	else if (cn->conv_engine == CONV_ENGINE_WNN) {
#ifndef DISABLE_WNN
		im_ja_wnn_select_candidate(cn, initial_selection);
		im_ja_wnn_update_preedit(cn);
#endif
	}
	candidate_window_hide(cn);
}
*/

void candidate_window_show(IMContextJA *cn, gint selected) {
	GtkWidget *candidate_item;
	GList *tmplist = NULL;
	gchar *tmpptr;
	gint cnt = 0;
	gint total = 0;

	IM_JA_DEBUG("show_candwin() with selected: %d\n", selected);

	if (cn->candidate_win != NULL) {
		gtk_widget_destroy(cn->candidate_win->window);
	}
	cn->candidate_win = g_new(CandidateWin, 1);

	if (cfg.candwin_style == CANDWIN_STYLE_TABLE) {
		GtkWidget *candidate_table;
		gint rows = 0;
		gint cols = 0;
		gint col, row;
		gdouble root = 0;
		tmplist = cn->candidate_list;
		// count the elements in the table
		while (tmplist != NULL) {
			total++;
			tmplist = g_list_next(tmplist);
		}
		//Calculate table size;
		root = sqrt((double) total);
		rows = cols = (gint) root;
		if (((gdouble) rows) < root) {
			rows++;
			if (rows * cols < total) cols++;
		}
		//IM_JA_DEBUG("Rows: %d, cols: %d\n", rows, cols);

		//Create the table
		cn->candidate_win->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_set_title(GTK_WINDOW(cn->candidate_win->window), _("Candidates"));
		gtk_window_set_modal(GTK_WINDOW(cn->candidate_win->window), TRUE);
		gtk_window_set_transient_for(GTK_WINDOW(cn->candidate_win->window), GTK_WINDOW(cn->toplevel_gtk));

		candidate_table = gtk_table_new(rows, cols, TRUE); 
		gtk_container_add(GTK_CONTAINER(cn->candidate_win->window), candidate_table);

   	tmplist = cn->candidate_list;
		col = row = 0;
		while (tmplist != NULL) {
			while (col < cols) {
				tmpptr = (gchar *)tmplist->data;
				if (tmplist->data != NULL) {
					GtkWidget *tmpwidget = gtk_button_new_with_label(tmplist->data);
					g_object_set_data(G_OBJECT(tmpwidget), "candidate-number", (gpointer) cnt);

					g_signal_connect(G_OBJECT(tmpwidget), "clicked", G_CALLBACK(candidate_window_set_selection), cn);

					gtk_table_attach(GTK_TABLE(candidate_table), tmpwidget, col, col + 1, row, row + 1,
													 (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
													 (GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);
					tmplist = g_list_next(tmplist);
					cnt++;
					col++;
					if (tmplist == NULL) break;
				}
			}
			col = 0;
			row++;
		}
	}
	else { //CANDWIN_STYLE_MENU
		cn->candidate_win->window = gtk_menu_new();
		tmplist = cn->candidate_list;
		while (tmplist != NULL) {
			tmpptr = (gchar *)tmplist->data;
			if (tmplist->data != NULL) {
				candidate_item = gtk_menu_item_new_with_label(tmplist->data);
				gtk_widget_show(candidate_item);
				g_object_set_data(G_OBJECT(candidate_item), "candidate-number", (gpointer) cnt);
				gtk_menu_shell_append(GTK_MENU_SHELL(cn->candidate_win->window), candidate_item);
				if (cnt == selected) 	gtk_menu_shell_select_item(GTK_MENU_SHELL(cn->candidate_win->window), candidate_item);
				
				g_signal_connect(GTK_OBJECT(candidate_item), "select", 
												 G_CALLBACK(candidate_window_change_selection), cn);
				g_signal_connect(GTK_OBJECT(candidate_item), "activate", 
												 G_CALLBACK(candidate_window_set_selection), cn);
				tmplist = g_list_next(tmplist);
				cnt++;
			}
			else {
				IM_JA_DEBUG("null received in GList\n");
				break;
			}
		}
		/*
		g_signal_connect_swapped(GTK_OBJECT(cn->candidate_win->window), "deactivate", G_CALLBACK(candidate_window_cancel_cb), cn);
		g_signal_connect_swapped(GTK_OBJECT(cn->candidate_win->window), "cancel", G_CALLBACK(candidate_window_cancel_cb), cn);
		g_signal_connect_swapped(GTK_OBJECT(cn->candidate_win->window), "selection-done", G_CALLBACK(candidate_window_selection_done), cn);
		*/
	}
	g_signal_connect(GTK_OBJECT(cn->candidate_win->window), "key_press_event", G_CALLBACK(key_press_cb), cn);
	g_signal_connect_swapped(GTK_OBJECT(cn->candidate_win->window), "destroy", G_CALLBACK(candidate_window_hide), cn);

	gtk_widget_show_all(cn->candidate_win->window);
	if (GTK_IS_WIDGET(cn->toplevel_gtk) == TRUE) g_object_set_data(G_OBJECT(cn->toplevel_gtk), "im-ja-candidate-window", cn->candidate_win);

	if (cfg.candwin_style == CANDWIN_STYLE_MENU) {
		gtk_menu_popup(GTK_MENU(cn->candidate_win->window), NULL, NULL, 
									 (GtkMenuPositionFunc) candidate_window_set_position, cn, 0, 
									 gtk_get_current_event_time());
	}

	status_window_hide(cn);
	kanjipad_hide(cn);
}

gboolean candidate_window_is_shown(IMContextJA* cn) {
	IM_JA_DEBUG("candidate_window_is_shown: %d\n", (int) g_object_get_data(G_OBJECT(cn->toplevel_gtk), "im-ja-candidate-window"));

	if (g_object_get_data(G_OBJECT(cn->toplevel_gtk), "im-ja-candidate-window") != NULL) return TRUE;
	return FALSE;
}


void candidate_window_hide(IMContextJA* cn) {
	IM_JA_DEBUG("candidate_window_hide()\n");
	if (GTK_IS_WIDGET(cn->toplevel_gtk) == TRUE) g_object_set_data(G_OBJECT(cn->toplevel_gtk), "im-ja-candidate-window", NULL);
	if (cn->candidate_win != NULL) {
		GtkWidget *tmpwidget = cn->candidate_win->window;
		cn->candidate_win->window = NULL;
		if (GTK_IS_WIDGET(tmpwidget) == TRUE) {
			gtk_widget_hide_all(tmpwidget);
			gtk_widget_destroy(tmpwidget);
		}
		g_free(cn->candidate_win);
		cn->candidate_win = NULL;
	}
	status_window_show(cn);
}

