/* -*- Mode: C; tab-width: 2; indent-tabs-mode: t; c-basic-offset: 2 -*- */
/* IM-JA Japanese Input Method Module for GTK-2.0
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 *
 * Authors: Botond Botyanszki <boti@rocketmail.com>
 *
 * Based on code by:
 *          Owen Taylor <otaylor@redhat.com>
 *          Yusuke Tabata <tee@kuis.kyoto-u.ac.jp>
 *          Yukihiro Nakai <ynakai@redhat.com>
 *          
 */
#include <config.h>

#include <string.h>
#include <stdio.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>
#include <gtk/gtkimmodule.h>
#include <gmodule.h>
#include <gconf/gconf-client.h>

#include "im-ja.h"
#include "common.h"
#include "error.h"
#include "conf.h"
#include "nls.h"
#include "romakana.h"
#include "statuswin.h"
#include "preeditwin.h"
#include "kanjipad.h"
#include "candwin.h"
#include "symbols.h"

#ifndef DISABLE_CANNA
#include "canna_rk.h"
#endif
#ifndef DISABLE_WNN
#include "wnn.h"
#endif


GType type_im_context_im_ja = 0;
extern GConfClient *gconf_client;

enum {
	STRING_COLUMN,
	N_COLUMNS
};

IMJAConfig cfg; 
gboolean kanjipad_focus_out_disabled;
gint notify_id;
gint requested_input_method;
gboolean im_changed_by_applet;

/* * * * * * * * * * * */

/* GtkIMContext API */
static void im_ja_class_init(GtkIMContextClass *class);
static void im_ja_class_finalize(GObject *obj);
static void im_ja_init(GtkIMContext *im_context);
static void im_ja_finalize(GtkIMContext *context);
static gboolean im_ja_filter_keypress(GtkIMContext *context, GdkEventKey *key);
static void im_ja_get_preedit_string(GtkIMContext *context, gchar **str, PangoAttrList **attrs, gint *cursor_pos);
static void im_ja_cursor_location_changed(GtkIMContext *context, GdkRectangle *area);
static void im_ja_set_client_window(GtkIMContext* context, GdkWindow *win);
static void im_ja_lost_focus(GtkIMContext *context);
static void im_ja_got_focus(GtkIMContext *context);
static void im_ja_set_use_preedit(GtkIMContext *context, gboolean use_preedit);
static void im_ja_on_reset(IMContextJA *context);

static gboolean init_conversion_engine(IMContextJA *context);
static void input_method_changed_cb(GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer data);
static void shutdown_conversion_engine(IMContextJA *cn);
static void im_ja_run_configurator();

/* * * * * * * * * * * */

GtkIMContext *im_ja_new(void) {
  IMContextJA *context;
  context = IM_CONTEXT_IM_JA(g_object_new(TYPE_IM_CONTEXT_IM_JA, NULL));
	IM_JA_DEBUG("im_ja_new() [%d]\n", (int) context);
  return GTK_IM_CONTEXT(context);
}

void im_ja_register_type(GTypeModule *module) {
	IM_JA_DEBUG("im_ja_register_type\n");
	static const GTypeInfo object_info = {
		sizeof (IMContextJAClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) im_ja_finalize,
		(GClassInitFunc) im_ja_class_init,
		NULL, //(GClassFinalizeFunc) im_ja_class_finalize,
		NULL,						/* class_data */
		sizeof (IMContextJA),
		0,
		(GInstanceInitFunc) im_ja_init,
	};

	type_im_context_im_ja = g_type_module_register_type(module, GTK_TYPE_IM_CONTEXT, "IMContextJA", &object_info, 0);
}

/* This is called only once for the application */
static void im_ja_class_init(GtkIMContextClass *class) {
	GtkIMContextClass *im_context_class = GTK_IM_CONTEXT_CLASS(class);
	GObjectClass *object_class = G_OBJECT_CLASS(class);
	IM_JA_DEBUG("im_ja_class_init\n");
	im_context_class->set_client_window = im_ja_set_client_window;
	im_context_class->filter_keypress = im_ja_filter_keypress;
	im_context_class->get_preedit_string = im_ja_get_preedit_string;
	im_context_class->set_cursor_location = im_ja_cursor_location_changed;
	im_context_class->focus_in = (void *)im_ja_got_focus;
	im_context_class->focus_out = (void *)im_ja_lost_focus;
	im_context_class->reset = (void*)im_ja_on_reset;
	im_context_class->set_use_preedit = (void *)im_ja_set_use_preedit;
	object_class->finalize = im_ja_class_finalize;

	gconf_client = NULL;

	if (im_ja_init_conf_handler() == FALSE) {
		im_ja_print_error(_("GConf initialization failed!"));
	}
	im_ja_get_gconf_client();
	cfg.im_ja_version = NULL;
	if (im_ja_load_conf(&cfg) == FALSE) {
		im_ja_print_error(_("Couldn't load settings!"));
	}

	if (cfg.im_ja_version == NULL) {
		if (im_ja_print_question(_("You have not yet configured im-ja.\nWould yo like to do so now?")) == TRUE) {
			gconf_client_set_string(gconf_client, GCONF_NAMESPACE"other/im_ja_version", VERSION, NULL);
			im_ja_run_configurator();
		}
		else gconf_client_set_string(gconf_client, GCONF_NAMESPACE"other/im_ja_version", VERSION, NULL);
	}
	else {
		if (strcmp(cfg.im_ja_version, VERSION) != 0) {
			if (im_ja_print_question(_("Would you like to update your im-ja settings?")) == TRUE) {
				gconf_client_set_string(gconf_client, GCONF_NAMESPACE"other/im_ja_version", VERSION, NULL);
				im_ja_run_configurator();
			}
			else gconf_client_set_string(gconf_client, GCONF_NAMESPACE"other/im_ja_version", VERSION, NULL);
		}
	}
	gconf_client_add_dir(gconf_client, GCONF_NAMESPACE"other", GCONF_CLIENT_PRELOAD_NONE, NULL);
	notify_id = gconf_client_notify_add(gconf_client,
																			GCONF_NAMESPACE"other/applet_input_method",
																			(GConfClientNotifyFunc) input_method_changed_cb,
																			NULL, NULL, NULL);
}

static void im_ja_class_finalize(GObject *obj) {
	IMContextJA *cn = IM_CONTEXT_IM_JA(obj);
  //GtkIMContextClass *im_context_class;

	IM_JA_DEBUG("im_ja_class_finalize() [%d]\n", (int)obj);

	cn->finalized = TRUE;
	
	preedit_window_destroy(cn);
	cn->preedit_win = NULL;
	status_window_destroy(cn);
	cn->status_win = NULL;
	shutdown_conversion_engine(cn);
	
	/*
	if (cn->preedit_buf != NULL) {
		g_free(cn->preedit_buf);
		cn->preedit_buf = NULL;
	}
	if (cn->preedit_insert != NULL) {
		g_free(cn->preedit_insert);
		cn->preedit_insert = NULL;
	}
	*/
	
	cn->client_gdk = NULL;
	cn->client_gtk = NULL;
	cn->toplevel_gtk = NULL;
	cn->toplevel_gdk = NULL;
	
	/* 
	im_context_class = GTK_IM_CONTEXT_GET_CLASS (cn);
	im_context_class->set_client_window = NULL;
	im_context_class->filter_keypress = NULL; 
	im_context_class->get_preedit_string = NULL;
	im_context_class->set_cursor_location = NULL;
	im_context_class->focus_in = NULL;
	im_context_class->focus_out = NULL;
	im_context_class->reset = NULL;
	im_context_class->set_use_preedit = NULL;
	*/
}

static void im_ja_finalize(GtkIMContext *context) {
	IM_JA_DEBUG("im_ja_finalize()  [%d]\n", (int)context);
	im_ja_finalize_conf_handler(&cfg);
	status_window_destroy_all();
	preedit_window_destroy_all();

}

/* Called for every input widget instance */
static void im_ja_init(GtkIMContext *context) {
	IMContextJA *cn = IM_CONTEXT_IM_JA(context);

	IM_JA_DEBUG("im_ja_init\n");
	cn->finalized = FALSE;
	cn->kanjipad = NULL;
	cn->preedit_buf = g_new0(guchar, BUFFERSIZE);
	cn->preedit_insert = g_new0(guchar, BUFFERSIZE);
	cn->client_gdk = NULL;
	cn->client_gtk = NULL;
	cn->status_win = NULL;
	cn->preedit_win = NULL;
	cn->preedit_reverse_start = 0;
	cn->preedit_reverse_end = 0;
	cn->cursor_ndx = -1;
	cn->show_first = FALSE;
	cn->toplevel_gtk = NULL;
	cn->toplevel_gdk = NULL;
	cn->preedit_win_on = cfg.preedit_window_on;
	cn->conv_engine_initialized = FALSE;
	im_ja_set_input_method(cn, cfg.startup_input_method);
}

static void im_ja_run_configurator() {
	static gchar *command[] = {IM_JA_BINDIR"/im-ja-conf"};
	GError *g_error = NULL;
	if (g_spawn_async(NULL, command, NULL, G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL,
										NULL, NULL, NULL, &g_error) == FALSE) {
		im_ja_print_error(g_error->message); 
	}
}

static void im_ja_set_use_preedit(GtkIMContext *context, gboolean use_preedit) {
	IMContextJA *cn = IM_CONTEXT_IM_JA(context);
	IM_JA_DEBUG("im_ja_set_use_preedit: %d\n", use_preedit);
	cn->preedit_win_on = use_preedit;
}

gboolean im_ja_filter_keypress(GtkIMContext *context, GdkEventKey *key) {
	gboolean retval = FALSE;
	IMContextJA *cn = IM_CONTEXT_IM_JA(context);

	if (key->type == GDK_KEY_RELEASE) return FALSE;

		
		IM_JA_DEBUG("GOT KEYEVENT. keyname: %s, keyval : %d,	state: %d\n", 
		gdk_keyval_name(key->keyval), key->keyval, key->state);
	

	if (ishotkey(key, START_CONFIGURATOR, &cfg) == TRUE) {
		im_ja_run_configurator();
		return TRUE;
	}
	if (ishotkey(key, NEXT_INPUT_MODE, &cfg) == TRUE) {
		im_ja_next_input_method(cn);
		return TRUE;
	}
	if (ishotkey(key, PREV_INPUT_MODE, &cfg) == TRUE) {
		im_ja_prev_input_method(cn);
		return TRUE;
	}
	if (ishotkey(key, RAW_INPUT_MODE, &cfg) == TRUE) {
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return TRUE;
	}
	if (ishotkey(key, HIRAGANA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already hiragana, then set to katakana */
		if (cn->input_method == IM_JA_HIRAGANA_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_HIRAGANA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, KATAKANA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already katakana, then set to half-widh katakana */
		if (cn->input_method == IM_JA_KATAKANA_INPUT) im_ja_set_input_method(cn, IM_JA_HALFKATA_INPUT);
		else im_ja_set_input_method(cn, IM_JA_KATAKANA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, HALFKATA_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already half-katakana, then set to full katakana */
		if (cn->input_method == IM_JA_HALFKATA_INPUT) im_ja_set_input_method(cn, IM_JA_KATAKANA_INPUT);
		else im_ja_set_input_method(cn, IM_JA_HALFKATA_INPUT);
		return TRUE;
	}
	if (ishotkey(key, ZENKAKU_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already zenkaku, then set to RAW */
		if (cn->input_method == IM_JA_ZENKAKU_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_ZENKAKU_INPUT);
		return TRUE;
	}
	if (ishotkey(key, CANNA_INPUT_MODE, &cfg) == TRUE) {
		im_ja_reset(cn);
		shutdown_conversion_engine(cn);
		cfg.default_conv_engine = CONV_ENGINE_CANNA;
		cn->conv_engine = cfg.default_conv_engine;
		return TRUE;
	}
	if (ishotkey(key, WNN_INPUT_MODE, &cfg) == TRUE) {
		im_ja_reset(cn);
		shutdown_conversion_engine(cn);
		cfg.default_conv_engine = CONV_ENGINE_WNN;
		cn->conv_engine = cfg.default_conv_engine;
		return TRUE;
	}
	if (ishotkey(key, KANJIPAD_INPUT_MODE, &cfg) == TRUE) {
		/* if it's already KANJIPAD, then set to RAW */
		if (cn->input_method == IM_JA_KANJIPAD_INPUT) im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		else im_ja_set_input_method(cn, IM_JA_KANJIPAD_INPUT);
		return TRUE;
	}
	if (ishotkey(key, TOGGLE_PREEDIT_WINDOW, &cfg) == TRUE) {
		cn->preedit_win_on = !cn->preedit_win_on;
		if (cn->preedit_win_on == TRUE) {
			status_window_hide(cn);
			preedit_window_show(cn);
		}
		else {
			preedit_window_hide(cn);
			status_window_show(cn);
		}
		return TRUE;
	}
	if (ishotkey(key, SYMBOL_INPUT, &cfg) == TRUE) {
		im_ja_symbol_table_show(cn);
	}
	if (ishotkey(key, UNICODE_INPUT, &cfg) == TRUE) {
		im_ja_unicode_entry_show(cn);
	}
	if (ishotkey(key, JIS_CODE_INPUT, &cfg) == TRUE) {
		im_ja_jiscode_entry_show(cn);
	}

	switch (cn->input_method) {
	case IM_JA_RAW_INPUT: //FIXME: Is there any way to do _real_ raw input instead of emulating it?
	case IM_JA_KANJIPAD_INPUT:
		if (strlen(cn->preedit_buf) > 0) {
			g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
			im_ja_reset(cn);
		}

		if (key->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK | GDK_MOD3_MASK | GDK_MOD4_MASK)) return FALSE; 

		if ((key->keyval >= GDK_exclam && key->keyval <= GDK_overline) ||
				(key->keyval >= GDK_KP_Space && key->keyval <= GDK_KP_9)) { 
			gchar utf8strg[7];
			int utf8strg_end;
			utf8strg_end = g_unichar_to_utf8(gdk_keyval_to_unicode(key->keyval), utf8strg);
			utf8strg[utf8strg_end] = 0;
			if (strlen(utf8strg) > 0) {
				if (((guchar)utf8strg[0] >= 0x20) && (utf8strg[0] != 0x7f)) g_signal_emit_by_name(cn, "commit", utf8strg);
			}
			else return FALSE;
			return TRUE;
		}
		if (key->keyval == GDK_space) {
			g_signal_emit_by_name(cn, "commit", " ");
			return TRUE;
		}
		return FALSE;
	case IM_JA_HIRAGANA_INPUT:
	case IM_JA_KATAKANA_INPUT:
	case IM_JA_HALFKATA_INPUT:
	case IM_JA_ZENKAKU_INPUT:
		if (cn->conv_engine_initialized == FALSE) {
			if (init_conversion_engine(cn) == FALSE) return FALSE; //Don't handle the key if we cannot init the engine.
		}
		if (cfg.default_conv_engine == CONV_ENGINE_CANNA) {
#ifndef DISABLE_CANNA
			retval = canna_rk_filter_keypress(cn, key);
#else
			IM_JA_DEBUG("This shouldn't happen. We are not supposed to be in canna mode if it's disabled\n");
#endif
		}
		else if (cfg.default_conv_engine == CONV_ENGINE_WNN) {
#ifndef DISABLE_WNN
			retval = im_ja_wnn_filter_keypress(cn, key);
#else
			IM_JA_DEBUG("This shouldn't happen. We are not supposed to be in wnn mode if it's disabled\n");
#endif
		break;
		}
	}

	/* ESC pressed: drop current pre-edit buffer if this keypress wasn't handled */
	if ((key->keyval == GDK_Escape) && (retval == FALSE)) {
		IM_JA_DEBUG("ESC: cancel preedit\n");
		IM_JA_DEBUG("preedit_insert: %d\n", strlen(cn->preedit_insert));
		IM_JA_DEBUG("preedit_buf: %d\n", strlen(cn->preedit_buf));
		if ((strlen(cn->preedit_buf) > 0) || (strlen(cn->preedit_insert) > 0)) { //Cancel preedit buffer
			im_ja_reset(cn);
#ifndef DISABLE_WNN
			if (cn->conv_engine == CONV_ENGINE_WNN) im_ja_wnn_reset(cn);
#endif
#ifndef DISABLE_CANNA
			if (cn->conv_engine == CONV_ENGINE_CANNA) canna_rk_reset(cn);
#endif
		}
		else {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		}
		return TRUE;
	}

	return retval;
}

void im_ja_set_preedit_string(IMContextJA *cn, gchar *str, gint reverse_start, gint reverse_end) {
	if (str == NULL) memset(cn->preedit_buf, 0, BUFFERSIZE);
	else g_strlcpy(cn->preedit_buf, str, BUFFERSIZE);
	cn->preedit_reverse_start = reverse_start;
	cn->preedit_reverse_end = reverse_end;
}

static void input_method_changed_cb(GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer data) {
	gint input_method = gconf_value_get_int(gconf_entry_get_value(entry));
	IM_JA_DEBUG("INPUT METHOD CHANGED BY APPLET\n");
	//update_status_label(applet);
	requested_input_method = input_method;
	im_changed_by_applet = TRUE;
	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"other/last_input_method", input_method, NULL);
}

void im_ja_get_preedit_string(GtkIMContext *ic, gchar **str, PangoAttrList **attrs, gint *cursor_pos) {
	IMContextJA *cn = IM_CONTEXT_IM_JA(ic);

	IM_JA_DEBUG("preedit_changed\n");

	if (attrs != NULL) {
		*attrs = pango_attr_list_new(); //Create new attrib list
	}

	if (g_utf8_validate(cn->preedit_buf, -1, NULL) != TRUE) {
		im_ja_print_error("utf8_validate failed for preedit string: %s\n", cn->preedit_buf);
	}
	else *str = g_strdup(cn->preedit_buf);
	
	if (*str == NULL || strlen(*str) == 0) {
		if ((cn->preedit_win_on == TRUE) && (cn->preedit_win != NULL)) {
			preedit_window_set_text(cn, *str);
			preedit_window_set_cursor_pos(cn, 0);
		}
		if (cursor_pos != NULL) *cursor_pos = 0;
		//else IM_JA_DEBUG("cursor_pos == NULL !!!\n");
		return;
	}

	PangoAttrList *preedit_win_attrs = pango_attr_list_new(); //Create new attrib list for the preedit window

	if (attrs != NULL) {
		/* Normal (and underlined) preedit string */
		PangoAttribute *backgr;
		PangoAttribute *foregr;
		PangoAttribute *underline;

		underline = pango_attr_underline_new(PANGO_UNDERLINE_SINGLE);
		underline->start_index = 0;
		underline->end_index = strlen(*str);
		pango_attr_list_insert(preedit_win_attrs, pango_attribute_copy(underline));
		pango_attr_list_insert(*attrs, underline);

		if (cfg.custom_preedit_n == TRUE) {
			foregr = pango_attr_foreground_new(cfg.preedit_colors[0].red, cfg.preedit_colors[0].green, cfg.preedit_colors[0].blue);
			backgr = pango_attr_background_new(cfg.preedit_colors[1].red, cfg.preedit_colors[1].green, cfg.preedit_colors[1].blue);
			backgr->start_index = 0;
			foregr->start_index = 0;
			backgr->end_index = strlen(*str);
			foregr->end_index = strlen(*str);
			pango_attr_list_insert(preedit_win_attrs, pango_attribute_copy(backgr));
			pango_attr_list_insert(preedit_win_attrs, pango_attribute_copy(foregr));
			pango_attr_list_insert(*attrs, backgr);
			pango_attr_list_insert(*attrs, foregr);
		}

		if (cn->preedit_reverse_start != cn->preedit_reverse_end) {
			/* Highlighted (==reverse) field */
			if (cfg.custom_preedit_hl == TRUE) {
				foregr = pango_attr_foreground_new(cfg.preedit_colors[2].red, cfg.preedit_colors[2].green, cfg.preedit_colors[2].blue);
				backgr = pango_attr_background_new(cfg.preedit_colors[3].red, cfg.preedit_colors[3].green, cfg.preedit_colors[3].blue);
			}
			else { /* Use the default */
				foregr = pango_attr_foreground_new(cn->original_colors[0].red, cn->original_colors[0].green, cn->original_colors[0].blue);
				backgr = pango_attr_background_new(cn->original_colors[1].red, cn->original_colors[1].green, cn->original_colors[1].blue);
			}
			backgr->start_index = cn->preedit_reverse_start;
			foregr->start_index = cn->preedit_reverse_start;
			backgr->end_index = cn->preedit_reverse_end;
			foregr->end_index = cn->preedit_reverse_end;
			IM_JA_DEBUG("reverse start: %d, end: %d\n", cn->preedit_reverse_start, cn->preedit_reverse_end);

			pango_attr_list_insert(preedit_win_attrs, pango_attribute_copy(backgr));
			pango_attr_list_insert(preedit_win_attrs, pango_attribute_copy(foregr));
			pango_attr_list_insert(*attrs, backgr);
			pango_attr_list_insert(*attrs, foregr);
		}
	}

	if (cursor_pos != NULL) {
		if (cn->conv_engine == CONV_ENGINE_WNN) {
			*cursor_pos = cn->cursor_wnn;
		}
		else if (cn->conv_engine == CONV_ENGINE_CANNA) {
			if (cn->cursor_ndx == -1) {
				*cursor_pos = strlen(*str);
			}
			else {
				//convert byte index into character index
				*cursor_pos = g_utf8_strlen(*str, cn->cursor_ndx);
			}
		}
		IM_JA_DEBUG("im_ja_get_preedit_string: cursor char pos: %d\n", *cursor_pos);
	}

	if (cn->preedit_win_on == TRUE) {
		gint preedit_cursor = 0;
		
		if (cn->conv_engine == CONV_ENGINE_WNN) {
			preedit_cursor = char_to_byte_pos(cn->preedit_buf, cn->cursor_wnn);
		}
		else if (cn->conv_engine == CONV_ENGINE_CANNA) {
			if (cn->cursor_ndx == -1) preedit_cursor = strlen(*str);
			else preedit_cursor = cn->cursor_ndx;
		}		
		
		preedit_window_set_cursor_pos(cn, preedit_cursor);
		preedit_window_set_text(cn, *str);
		preedit_window_set_attributes(cn, preedit_win_attrs);
	}
}

/*
static void im_ja_cleanup(IMContextJA *cn) {
  GtkIMContextClass *im_context_class;

	IM_JA_DEBUG("im_ja_cleanup()\n");
		
	preedit_window_destroy(cn);
	cn->preedit_win = NULL;
	status_window_destroy(cn);
	cn->status_win = NULL;
	shutdown_conversion_engine(cn);
	IM_JA_DEBUG("im_ja_cleanup() finished\n");
	if (cn->preedit_buf != NULL) {
		g_free(cn->preedit_buf);
		cn->preedit_buf = NULL;
	}
	if (cn->preedit_insert != NULL) {
		g_free(cn->preedit_insert);
		cn->preedit_insert = NULL;
	}
	cn->client_gdk = NULL;
	cn->client_gtk = NULL;
	cn->toplevel_gtk = NULL;
	cn->toplevel_gdk = NULL;

  im_context_class = GTK_IM_CONTEXT_GET_CLASS (cn);
	im_context_class->set_client_window = NULL;
	im_context_class->filter_keypress = NULL;
	im_context_class->get_preedit_string = NULL;
	im_context_class->set_cursor_location = NULL;
	im_context_class->focus_in = NULL;
	im_context_class->focus_out = NULL;
	im_context_class->reset = NULL;
	im_context_class->set_use_preedit = NULL;
}
*/

static void im_ja_set_client_window(GtkIMContext *context, GdkWindow *win) {
	IMContextJA* cn = IM_CONTEXT_IM_JA(context);
	IM_JA_DEBUG("im_ja_set_client_window [context:%d, window:%d]\n", (int) context, (int) win);
	if (cn->finalized == TRUE) {
		IM_JA_DEBUG("*ERROR* IMContext is already finalized.\n");
		return;
	}
	cn->client_gdk = win;
	if (win == NULL) {
		/*
			preedit_window_destroy(cn);
			cn->preedit_win = NULL;
			status_window_destroy(cn);
			cn->status_win = NULL;
			shutdown_conversion_engine(cn);
		*/
		/*
			if (cn->preedit_buf != NULL) {
				g_free(cn->preedit_buf);
				cn->preedit_buf = NULL;
			}
			if (cn->preedit_insert != NULL) {
				g_free(cn->preedit_insert);
				cn->preedit_insert = NULL;
			}
		*/
		cn->client_gdk = NULL;
		cn->client_gtk = NULL;
		cn->toplevel_gtk = NULL;
		cn->toplevel_gdk = NULL;
		return; 
	}
	if (GDK_IS_WINDOW(cn->client_gdk) == TRUE) cn->toplevel_gdk = gdk_window_get_toplevel(cn->client_gdk);
	gdk_window_get_user_data(cn->toplevel_gdk, (gpointer *)&cn->toplevel_gtk);
	gdk_window_get_user_data(cn->client_gdk, (gpointer *)&cn->client_gtk);
	/*
	cn->client_win_destroy_handler = g_signal_connect_swapped(G_OBJECT(client_gtk), "destroy",
																														G_CALLBACK(im_ja_cleanup), cn);
	*/

	/* Maybe this is better ?
	cn->original_colors[1] = cn->client_gtk->style->base[GTK_STATE_SELECTED];
	cn->original_colors[0] = cn->client_gtk->style->text[GTK_STATE_SELECTED];
	*/
	cn->original_colors[0] = cn->client_gtk->style->base[GTK_STATE_NORMAL];
	cn->original_colors[1] = cn->client_gtk->style->text[GTK_STATE_NORMAL];
  //g_signal_connect_swapped(cn->toplevel_gtk, "destroy", G_CALLBACK(im_ja_finalize), cn);
}

void im_ja_free_candidate_list(IMContextJA *cn) {
	GList *tmplist;
	
	tmplist = cn->candidate_list;
	if (tmplist == NULL) return;

	while (tmplist != NULL) {
		g_free(tmplist->data);
		tmplist = g_list_next(tmplist);
	}
	g_list_free(cn->candidate_list);
	cn->candidate_list = NULL;
}

void im_ja_reset(IMContextJA *cn) {
	IM_JA_DEBUG("im_ja_reset()\n");
	*cn->preedit_insert = 0;

	if (cn->preedit_buf == NULL) cn->preedit_buf = g_new0(guchar, BUFFERSIZE);
	if (cn->preedit_insert == NULL) cn->preedit_insert = g_new0(guchar, BUFFERSIZE);

	memset(cn->preedit_buf, 0, BUFFERSIZE); //This makes gnome-terminal segfault;
	cn->preedit_reverse_start = 0;
	cn->preedit_reverse_end = 0;
	cn->cand_stat = 0;
	cn->cursor_ndx = -1;
	cn->cursor_wnn = 0;
	g_signal_emit_by_name(cn, "preedit_changed");
	g_signal_emit_by_name(cn, "preedit_end");
	cn->candwin_pos_offset_x = cn->cursor_pos_offset_x; // Cursor position does not always change
	cn->candwin_pos_offset_y = cn->cursor_pos_offset_y; // so we need to update manually
}


void im_ja_grab_add(GtkWidget *widget, GdkEvent *event, IMContextJA* cn) {
	gint success = -1; 
	IM_JA_DEBUG("add grab:");
	success = gdk_keyboard_grab(cn->toplevel_gdk, FALSE, GDK_CURRENT_TIME);
	if (success == GDK_GRAB_SUCCESS) IM_JA_DEBUG(" success\n");
	else IM_JA_DEBUG(" failed\n");
}

void im_ja_grab_remove(GtkWidget *widget, GdkEvent *event, IMContextJA* cn) {
	IM_JA_DEBUG("remove grab\n");
	gdk_keyboard_ungrab(GDK_CURRENT_TIME);
}

static void im_ja_cursor_location_changed(GtkIMContext *context, GdkRectangle *area) {
	IMContextJA* cn = IM_CONTEXT_IM_JA(context);
	gint x, y, area_h;

	if (cn->finalized == TRUE) {
		IM_JA_DEBUG("*ERROR* IMContext is already finalized.\n");
		return;
	}

	//Bug in GTK?:	width & height is sometimes (on first set_location) a huge_number
	area_h = area->height;
	if ((area->width > 1000) || (area->width < 0) ||
			(area->height > 1000) || (area->height < 0)) {
		area_h = 12;
	}
	x = area->x;
	y = area->y + area_h/2;

	if (((strlen(cn->preedit_buf) == 0) || (cn->update_preedit_pos == TRUE)) 
			&& ((cn->preedit_win_pos_offset_x != x) || (cn->preedit_win_pos_offset_y != y))) {
		cn->preedit_win_pos_offset_x = x;
		cn->preedit_win_pos_offset_y = y;
		cn->update_preedit_pos = FALSE;
		if (cn->preedit_win_on == TRUE) preedit_window_update_location(cn);
	}			
	
	//IM_JA_DEBUG("im_ja_cursor_location_changed: %d, %d\n", x, y);
	if ((cn->cursor_pos_offset_x != x) || (cn->cursor_pos_offset_y != y)) {
		if ((strlen(cn->preedit_buf) == 0) || (cn->update_candwin_pos == TRUE)) {
			cn->candwin_pos_offset_x = x;
			cn->candwin_pos_offset_y = y;
			cn->update_candwin_pos = FALSE;
			//IM_JA_DEBUG("candwin cursor_location_changed(%d, %d)\n", x, y);
		}
		cn->cursor_pos_offset_x = x;
		cn->cursor_pos_offset_y = y;
		//IM_JA_DEBUG("cursor_location_changed(%d, %d)\n", x, y);
		kanjipad_set_location(cn);
		if (cn->preedit_win_on == TRUE) preedit_window_update_location(cn);
		else status_window_update_location(cn);
	}
}

static void im_ja_lost_focus(GtkIMContext *context) {
	IMContextJA* cn = IM_CONTEXT_IM_JA(context);
	
	IM_JA_DEBUG("im-ja lost focus\n");

	preedit_window_hide(cn);
	status_window_hide(cn);
		
	if (kanjipad_focus_out_disabled != TRUE) {
		kanjipad_hide(cn);
	}
}

static void im_ja_got_focus(GtkIMContext *context) {
	IMContextJA* cn = IM_CONTEXT_IM_JA(context);
	gint applet_im;

	IM_JA_DEBUG("im-ja got focus\n");
	if (im_changed_by_applet == TRUE) {
		im_changed_by_applet = FALSE;
		im_ja_set_input_method(cn, requested_input_method);
	}

	//Check if applet has the same input method
	applet_im = gconf_client_get_int(gconf_client, GCONF_NAMESPACE"other/last_input_method", NULL);
	if (applet_im != cn->input_method) {
		gconf_client_set_int(gconf_client, GCONF_NAMESPACE"other/last_input_method", cn->input_method, NULL);
	}

	//	if (cn->status_window != NULL) gtk_grab_remove(cn->status_window);
	if (cn->status_win != NULL) cn->status_win->can_hide = TRUE;
	if (cn->preedit_win != NULL) cn->preedit_win->can_hide = TRUE;
	if (cn->show_first == FALSE) {
		cn->show_first = TRUE;
	}
	if (cn->preedit_win_on == TRUE) {
		preedit_window_show(cn);
	}
	else {
		status_window_show(cn);
	}
	kanjipad_set_location(cn);
	kanjipad_focus_out_disabled = FALSE;
	if (cn->input_method == IM_JA_KANJIPAD_INPUT) {
		kanjipad_show(cn);
	}
}

static void im_ja_on_reset(IMContextJA *cn) {
	IM_JA_DEBUG("im_ja_on_reset()\n");

	cn->preedit_win_pos_offset_x = cn->cursor_pos_offset_x;
	cn->preedit_win_pos_offset_y = cn->cursor_pos_offset_y;

	if (cn->preedit_buf == NULL) return;
	if (strlen(cn->preedit_buf) > 0) {
		g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
		im_ja_reset(cn);

#ifndef DISABLE_WNN
		if (cn->conv_engine == CONV_ENGINE_WNN) im_ja_wnn_reset(cn);
#endif
#ifndef DISABLE_CANNA
		if (cn->conv_engine == CONV_ENGINE_CANNA) canna_rk_reset(cn);
#endif
	}
}

static void shutdown_conversion_engine(IMContextJA *cn) {
	IM_JA_DEBUG("shutdown_conversion_engine()\n");
	if (cn->conv_engine_initialized == TRUE) {
		cn->conv_engine_initialized = FALSE;
		switch (cn->conv_engine) {
		case CONV_ENGINE_WNN:
#ifndef DISABLE_WNN
			im_ja_wnn_shutdown(cn);
			break;
#endif
		case CONV_ENGINE_CANNA:
#ifndef DISABLE_CANNA
			IM_JA_DEBUG("FIXME: finalize canna connections.\n");
#endif
			break;
		}
	}
}

static gboolean init_conversion_engine(IMContextJA *cn) {
	IM_JA_DEBUG("init_conversion_engine\n");

	if (cn->conv_engine_initialized == TRUE) return TRUE;
	if (cn->conv_engine == CONV_ENGINE_WNN) {
#ifndef DISABLE_WNN
		if (im_ja_wnn_init(cn) != TRUE) {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		} 
		else cn->conv_engine_initialized = TRUE;
#else
		im_ja_print_error(_("Wnn support is disabled. If you want wnn support, recompile im-ja."));
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return FALSE;
#endif
	}
	else if (cn->conv_engine == CONV_ENGINE_CANNA) {
#ifndef DISABLE_CANNA
		if (canna_rk_init(cn) != TRUE) {
			im_ja_print_error(_("canna init failed.\nPlease check your settings and make sure the canna server is running!")); 
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		}
		else cn->conv_engine_initialized = TRUE;
#else
		im_ja_print_error(_("Canna support is disabled. If you want canna support, recompile im-ja."));
		im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
		return FALSE;
#endif
	}
	return TRUE;
}

gboolean im_ja_set_input_method(IMContextJA *cn, int input_method) {
	gint old_input_method = cn->input_method;
	IM_JA_DEBUG("setting input method: %d\n", input_method);

	cn->input_method = input_method;
	candidate_window_hide(cn);

	gconf_client_set_int(gconf_client, GCONF_NAMESPACE"other/last_input_method", input_method, NULL);

	if (cn->input_method != IM_JA_KANJIPAD_INPUT) kanjipad_hide(cn);
	switch (cn->input_method) {
	case IM_JA_KANJIPAD_INPUT:
		if (cfg.kanjipad_enabled == TRUE) {
			if (input_method != old_input_method) {
				shutdown_conversion_engine(cn);
				if (strlen(cn->preedit_buf) > 0) g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
				im_ja_reset(cn);
			}
			cn->conv_engine = CONV_ENGINE_KANJIPAD;
			status_window_hide(cn);
			preedit_window_hide(cn);
			kanjipad_show(cn);
		}
		else {
			im_ja_set_input_method(cn, IM_JA_RAW_INPUT);
			return FALSE;
		}
		break;
	case IM_JA_RAW_INPUT:
		if (input_method != old_input_method) {
			im_ja_on_reset(cn);
		}
		status_window_hide(cn);
		preedit_window_hide(cn);
		break;
	default: 
		cn->conv_engine = cfg.default_conv_engine;
		status_window_show(cn);
	}
	return TRUE;
}

void im_ja_commit(IMContextJA *cn) {
	IM_JA_DEBUG("im_ja_commit()\n");

	switch (cn->conv_engine) {
	case CONV_ENGINE_WNN:
#ifndef DISABLE_WNN
		im_ja_wnn_commit(cn);
#endif
		break;
		/*
		case CONV_ENGINE_CANNA:
		#ifndef DISABLE_CANNA
		
		#endif
		*/
	}

	g_signal_emit_by_name(cn, "commit", cn->preedit_buf);
	im_ja_reset(cn);
	cn->update_preedit_pos = TRUE;
	if (cn->preedit_win != NULL) {
		gtk_window_resize(GTK_WINDOW(cn->preedit_win->window), 140, 25);
	}
}

void im_ja_next_input_method(IMContextJA *cn) {
	gint tmp;
	tmp = cn->input_method;
	tmp++;
	if ((tmp == IM_JA_KANJIPAD_INPUT) && (cfg.kanjipad_enabled == FALSE)) tmp++;
	if (tmp >= IM_JA_INPUT_METHODS_TOTAL) tmp = IM_JA_RAW_INPUT;
	im_ja_set_input_method(cn, tmp);
}

void im_ja_prev_input_method(IMContextJA *cn) {
	gint tmp;
	tmp = cn->input_method;
	tmp--;
	if ((tmp == IM_JA_KANJIPAD_INPUT) && (cfg.kanjipad_enabled == FALSE)) tmp--;
	if (tmp < IM_JA_RAW_INPUT) tmp = IM_JA_INPUT_METHODS_TOTAL - 1;
	im_ja_set_input_method(cn, tmp);
}
 
void im_ja_input_utf8(gchar *strg, IMContextJA *cn) {
	 
	if (strg == NULL) return;
	if ((cn->input_method != RAW_INPUT_MODE) || (cn->input_method != KANJIPAD_INPUT_MODE)) im_ja_on_reset(cn);
	IM_JA_DEBUG("im_ja_input_utf8(%s)\n", strg);
	g_signal_emit_by_name(cn, "commit", strg);
	
}
